from OFS import SimpleItem
import threading, time, xmlrpclib, sys
from Helpers import Helpers
from Globals import HTMLFile
import Globals
import ConflictResolvingTools

def manage_addZQueuePacket(self, id, topic, originatorId, recipients, messages=[], locked=0):
    '''Add a ZQueueTopic to the self folder.'''
    newOb = ZQueuePacket(id, topic, originatorId, recipients, messages, locked)
    self._setObject(id, newOb)
    return newOb

Globals.zQueuePacketThreadLocks = {}
Globals.zQueuePacketThreadLocksLock = threading.Lock()

class ZQueuePacket(SimpleItem.SimpleItem):
    '''ZQueue Packet Object'''
    meta_type = "ZQueuePacket"

    def __init__(self, id, topic, originatorId, recipients, messages=[], locked=0):
        '''Initialiser
        id - a unique id for this packet
        topic - the topic the messages belong to
        originatorId - the originator of this packet
        recipients - the recipients left for this packet'''
        self.id = id
        self.topicId = topic
        self.originatorId = originatorId
        self.messages = messages
        self.recipients = recipients
        self.locked = locked #editing lock
        self.lastEditTime = time.time()
        self.messagesAreDeliveredLocally = 0


    def __del__(self):
        '''Remove locks from store'''
        Globals.zQueuePacketThreadLocksLock.acquire()
        try:
            if Globals.zQueuePacketThreadLocks.has_key(self.id):
                del Globals.zQueuePacketThreadLocks[self.id] #remove our lock from the store
        finally:
            Globals.zQueuePacketThreadLocksLock.release()

    def getMessagesAreDeliveredLocally(self):
        '''Return the list of messages already delivered locally.'''
        return self.messagesAreDeliveredLocally

    def setMessagesAreDeliveredLocally(self):
        '''Sets the messagesAreDeliveredLocally to true.'''
        self.messagesAreDeliveredLocally = 1

    def getTimeSinceLastEdit(self):
        '''Returns the time since the last edit.'''
        return time.time() - self.lastEditTime

    def getThreadLock(self):
        '''Locks this object for multi threaded purposes.'''
        Globals.zQueuePacketThreadLocksLock.acquire()
        try:
            if not Globals.zQueuePacketThreadLocks.has_key(self.id):
                Globals.zQueuePacketThreadLocks[self.id] = threading.Lock() #threading lock
            threadLock = Globals.zQueuePacketThreadLocks[self.id]
        finally:
            Globals.zQueuePacketThreadLocksLock.release()
            threadLock.acquire()


    def releaseThreadLock(self):
        '''Unlocks this object for multi threaded purposes.'''
        Globals.zQueuePacketThreadLocksLock.acquire()
        try:
            if not Globals.zQueuePacketThreadLocks.has_key(self.id):
                Globals.zQueuePacketThreadLocks[self.id] = threading.Lock() #threading lock
            threadLock = Globals.zQueuePacketThreadLocks[self.id]
        finally:
            Globals.zQueuePacketThreadLocksLock.release()
            threadLock.release()

    def getLocked(self):
        '''Returns if this packet is locked'''
        return self.locked

    def setLocked(self, locked):
        '''Sets the locked status of this packet.'''
        print "Packet %s setLocked to %s" %(self.id, locked)
        self.locked = locked
        self._p_changed = 1

    def setAllDelivered(self):
        '''Override method used when root originator notifies that this packed is delivered.'''
        self.recipients = {}
        self._p_changed = 1

    def getRecipients(self):
        '''Returns the list of recipients in this packet.'''
        return self.recipients

    def getTopicId(self):
        '''The topic of this packet.'''
        return self.topicId

    def removeRecipient(self, recipient):
        '''Remove recipient and update allDelivered.
        Also notify the originator if there is one.'''
        originatorId = self.getOriginatorId()
        if originatorId is not None:
            #we are not the originator, so we need to notify the originator of the delivery.
            try:
                address = self.aq_parent.getRemoteNodes()[originatorId]
                if address:
                    s = xmlrpclib.Server(address)
                    delivered = s.packetDelivered(self.id, recipient)
            except:
                tb = Helpers.GetTraceback()
                print "Node %s" %(self.aq_parent.id)
                print tb
                #XXX Catch more specific??
                pass
        #remove the recipient from the packet as it has been delivered.
        del self.recipients[recipient]
        self._p_changed = 1

    def addMessage(self, messageId, message):
        '''Add a message to the packet'''
        self.messages.append((messageId,message))
        self.lastEditTime = time.time()
        self._p_changed = 1

    def getMessages(self):
        '''Return the messages in the packet'''
        return self.messages

    def getNumberOfMessages(self):
        '''Return the number of messages in the packet.'''
        return len(self.messages)

    def getOriginatorId(self):
        '''Return the originator's id.'''
        return self.originatorId

    def index_html(self):
        '''Render the contents of this packet.'''
        packetId = self.id
        packetTopic = self.getTopicId()

        originator = self.originatorId
        recipientsList = []
        for recipient in self.recipients.keys():
            recipientsList.append('''Node %s<br>''' %(recipient))
        recipientsList = '\n'.join(recipientsList)

        packetContents = []
        if self.getLocked():
            lockedStatus = 'is locked'
        else:
            lockedStatus = 'is unlocked'
        for messageId, message in self.messages:
            packetContents.append("%s : %s" %(messageId, message))

        packetContents = '<br>\n'.join(packetContents)
        return HTMLFile('dtml/packet', globals())(client=self, mapping=locals())


    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Conflict resolver.'''
        try:
            #set up a handler with a default handler that raises an error if the values are different
            handler = ConflictResolvingTools.KeyConflictHandler(ConflictResolvingTools.errorIfDifferent)
            handler.addHandleableKey('_objects', ConflictResolvingTools.tupleMerge)
            handler.addHandleableKey('lastEditTime', ConflictResolvingTools.returnLargest)
            handler.addHandleableKey('messagesAreDeliveredLocally', ConflictResolvingTools.returnSmallest)
            handler.addHandleableKey('locked', ConflictResolvingTools.returnSmallest)
            handler.addHandleableKey('messages', ConflictResolvingTools.listMerge)
            handler.addHandleableKey('recipients', ConflictResolvingTools.dictMerge)
            retState = ConflictResolvingTools.mergeDiffs(oldState, savedState, newState, handler)
            #make sure that there are keys for every id in _objects.  Remove those that are not relevant
            ConflictResolvingTools.synchroniseObjects(retState)
            return retState
        except:
            tb = Helpers.GetTraceback()
            print "!" * 120
            print "Error resolving conflict for Packet"
            print tb
            return None        
      
##        try:
##            print "PACKET CONFLICT RESOLVER"
##            if __debug__:
##                assertList = ['id','lastEditTime','locked','messages','messagesAreDeliveredLocally',
##                    'originatorId','recipients','topicId']
##
##                for key in s.keys():
##                    assert key in assertList, "This handler is not designed to handle key %s" %key
##
##            assert s['id'] == n['id']
##            assert s['originatorId'] == n['originatorId']
##            assert s['topicId'] == n['topicId']
##
##            #use the latest time
##            s['lastEditTime'] = n['lastEditTime'] > s['lastEditTime'] or s['lastEditTime']
##            #locked == 0 gets precedence
##            s['locked'] = s['locked'] and n['locked']
##            #messagesAreDeliveredLocally == 0 gets precedence
##            s['messagesAreDeliveredLocally'] = s['messagesAreDeliveredLocally'] and n['messagesAreDeliveredLocally']
##
##            #combine messages without making duplicates
##            Utils.combineListsUniquely(s['messages'],n['messages'])
##
##            #combine the recipients and check the entries that are already in both to be the same.
##            for key in n['recipients'].keys():
##                if s['recipients'].has_key(key):
##                    assert s['recipients'][key] == n['recipients'][key], 'Recipient addresses do not match!'
##                else:
##                    s['recipients'][key] = n['recipients'][key]
##
##            return s
##        except:
##            if hasattr(sys,'breakpoint'):
##                sys.breakpoint()
##            tb = Helpers.GetTraceback()
##            print tb
##            raise
