'''Regression testing module for ZMessageQueue'''

import Node, xmlrpclib, unittest, sys, traceback, cStringIO, time
from Globals import HTMLFile
import Globals
from OFS import Folder
from Helpers import Helpers

def manage_addZQueueTest(self, id='ZQueueTest'):
    '''Factory method to create a ZQeueTest
    id : Zope Id (default = ZQueueNode)'''
    self._setObject(id, ZQueueTest(id))
    return 1

class ZQueueTest(Folder.Folder):
    '''Class to do regression testing on the Node components.'''
    meta_type = "ZQueueTest"
    
    def __init__(self, id):
        '''Initialiser.'''
        self.id = id
        self.deleteTestObjectsAfterTest = 1
        
    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Confilct resolver.'''
        print "!" * 120
        print " NO CONFLICT RESOLVER FOR        ZQUEUETEST"
        print "!" * 120
        return None
        
        
    def setDeleteObjectsAfterTest(self, value=1):
        '''Set the flag to delete test objects after test is run.'''
        self.deleteTestObjectsAfterTest = value

    def renderTest(self, testCases, deleteObjects=0):
        '''Do the test and render the results.
        testCases : a list of the test cases to run
        deleteObjects : delete all the test objects that were created.'''
        t = time.time()
        print "deleteObjectsAfterTest = %s" %deleteObjects
        
        suite = unittest.TestSuite()
        
        for testCase in testCases:
            suite.addTest(MessageQueueTest(testCase, self, deleteObjects))
        
        results = unittest.TestResult()
        suite.run(results)
        if hasattr(sys,'breakpoint'):
            sys.breakpoint()
        numErrors = len(results.errors)
        numFailures = len(results.failures)

        #build the errors list
        errorsList = []
        for ob, tb in results.errors:
            if type(tb) == type(()):
                tbFile=cStringIO.StringIO()
                traceback.print_exception(tb[0],tb[1],tb[2],None, tbFile)
                tbFile.seek(0)
                tb = tbFile.read()
            errorsList.append(tb)

        #build the failures list
        failuresList = []
        for ob, tb in results.failures:
            if type(tb) == type(()):
                tbFile=cStringIO.StringIO()
                traceback.print_exception(tb[0],tb[1],tb[2],None, tbFile)
                tbFile.seek(0)
                tb = tbFile.read()
            failuresList.append(tb)

        timeTaken = time.time() - t

        outputOb = HTMLFile('dtml/renderTest',globals())
        return outputOb(client=self, mapping=locals())
    
    def index_html(self):
        '''Default view.'''
        #Get all the possible tests
        testCases = [name for name in MessageQueueTest.__dict__.keys() if name[0:4] == 'test']
        testCases = [[name, MessageQueueTest.__dict__[name].func_doc.replace('\n','<br>')] 
                        for name in testCases]
        outputOb = HTMLFile('dtml/test',globals())
        return outputOb(client=self, mapping=locals())

class MessageQueueTest(unittest.TestCase):
    '''Class encapsulating the test.'''
    def __init__(self, methodName, context, deleteObjectsAfterTest):
        '''Initialiser.'''
        unittest.TestCase.__init__(self, methodName)
        self.context = context
        self.deleteObjectsAfterTest = deleteObjectsAfterTest
    
    def setUp(self):
        '''Init the class with a pointer to the context to work in.'''
        #this must be set to a valid Zope context!
        #see below
        print "Set Up"
        self.server = xmlrpclib.Server(self.context.absolute_url())
        self.messages = []
        self.createTestNodes()
    
    def tearDown(self):
        '''Clean up everything.'''
        print "tearDown, deleteObjectsAfterTest = %s" %self.deleteObjectsAfterTest
        if self.deleteObjectsAfterTest != '0' and self.deleteObjectsAfterTest:
            print "Deleting objects."
            try:
                self.server.TestNode1.stopDeliveryHandler(0)
                self.server.TestNode2.stopDeliveryHandler(0)
            finally:
                self.context.manage_delObjects(['TestNode1','TestNode2'])
    
    def createTestNodes(self):
        '''Create two test nodes in the current context.'''
        obs = self.context.objectValues()
        for ob in obs:
            try:
                ob.stopDeliveryHandler()
            except:
                pass
        
        self.context.manage_delObjects(self.context.objectIds())
        #Create the nodes
        Node.manage_addZQueueNode(self.context, id='TestNode1', nodeId='TestNode1')
        Node.manage_addZQueueNode(self.context, id='TestNode2', nodeId='TestNode2')
        get_transaction().commit()
        
        #Add references of each other between the two nodes.
        self.server.TestNode1.addRemoteNode('TestNode2', str("%s/TestNode2" %self.context.absolute_url()), 0)
        self.server.TestNode2.addRemoteNode('TestNode1', str("%s/TestNode1" %self.context.absolute_url()), 0)

        #Start the node handlers
        self.server.TestNode1.startDeliveryHandler(0)
        self.server.TestNode2.startDeliveryHandler(0)

    def publishMessages(self, numberOfMessages = 10):
        '''Publish 10 messages to TestNode1'''
        self.messagesToSend = ["This is message %s" %i for i in range(numberOfMessages)]
        self.messagesToSend.sort()
        for message in self.messagesToSend:
            self.server.TestNode1.publishMessageToTopic('Topic1',message)
        
    def subscribeToMessages(self):
        '''Subscribes us to Topic1'''
        self.server.TestNode2.subscribeToTopic('Topic1','TestApp')
        
    def getMessages(self):
        '''Retrieve messages from Topic1 on TestNode2'''
        #get the messages
        messages = self.server.TestNode2.readMessagesFromTopic('Topic1','TestApp')
        if messages:
            self.messages.extend([message for messageId, message in messages])
            
        #acknowledge the messages.
        messageIds = [messageId for messageId, message in messages]
        self.server.TestNode2.ackMessagesFromTopic('Topic1', 'TestApp', messageIds)
        
    def testPublishGetMessages(self):
        '''Test publishing and receiving of messages.
        This test times out after 20 seconds.'''
        t = time.time()
        #do a getMessages to ensure we are subscribed
        self.subscribeToMessages()
        numberOfMessages = 10
        self.publishMessages(numberOfMessages)
        t = time.time()
        timeOut = 20
        while time.time() - t < timeOut and len(self.messages) < numberOfMessages:
            self.getMessages()
            if len(self.messages) < numberOfMessages:
                time.sleep(1)
        self.messages.sort()
        self.failUnless(self.messages == self.messagesToSend,
            "Messages sent:\n%s\n is not equal to messages retrieved:\n%s" %(self.messagesToSend, self.messages))

        self.timeTaken = time.time() - t

            
    def testHeavyLoad(self):
        '''Test publishing and receiving of messages under heavy load.
         Publishes 1000 messages in a thread while simultatiously receiving them.
         Handy to see if conflict resolution works ok.
         This test times out after 240 seconds.'''
        t = time.time()
        self.messages = []
        numberOfMessages = 1000
        import thread
        #publish a 1000 messages in a new thread
        thread.start_new_thread(self.publishMessages,(numberOfMessages,))
        self.subscribeToMessages()
        timeOut = 240
        t = time.time()
        while time.time() - t < timeOut and len(self.messages) < numberOfMessages:
            self.getMessages()
            if len(self.messages) < numberOfMessages:
                time.sleep(1)
        self.messages.sort()
        self.failUnless(self.messages == self.messagesToSend,
            "Sent %s messages and received %s.\n Messages sent:\n%s \n is not equal to messages retrieved:\n%s" %(
                len(self.messagesToSend), len(self.messages),self.messagesToSend, self.messages))
        
        self.timeTaken = time.time() - t
    